//--------------------------------------------
// PolylineArc.cs (c) 2008 by Charles Petzold
//--------------------------------------------
using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Media;
using System.Windows.Shapes;

namespace ArcSegmentMathDemo
{
    public class PolylineArc : Shape
    {
        PathGeometry pathGeo = new PathGeometry();
        PathFigure pathFig = new PathFigure();
        PolyLineSegment polylineSeg = new PolyLineSegment();

        // Dependency properties
        public static readonly DependencyProperty StartPointProperty = 
            PathFigure.StartPointProperty.AddOwner(typeof(PolylineArc),
                new FrameworkPropertyMetadata(new Point(0, 0), 
                        PropertyChangedCallback)); 

        public static readonly DependencyProperty PointProperty = 
            ArcSegment.PointProperty.AddOwner(typeof(PolylineArc),
                new FrameworkPropertyMetadata(new Point(0, 0), 
                        PropertyChangedCallback)); 

        public static readonly DependencyProperty SizeProperty = 
            ArcSegment.SizeProperty.AddOwner(typeof(PolylineArc),
                new FrameworkPropertyMetadata(new Size(0, 0),
                        PropertyChangedCallback)); 

        public static readonly DependencyProperty SweepDirectionProperty = 
            ArcSegment.SweepDirectionProperty.AddOwner(typeof(PolylineArc),
                new FrameworkPropertyMetadata(SweepDirection.Counterclockwise,
                        PropertyChangedCallback)); 

        public static readonly DependencyProperty IsLargeArcProperty = 
            ArcSegment.IsLargeArcProperty.AddOwner(typeof(PolylineArc),
                new FrameworkPropertyMetadata(false,
                        PropertyChangedCallback)); 

        public static readonly DependencyProperty RotationAngleProperty = 
            ArcSegment.RotationAngleProperty.AddOwner(typeof(PolylineArc),
                new FrameworkPropertyMetadata(0.0,
                        PropertyChangedCallback)); 

        // Property-changed handler
        static void PropertyChangedCallback(DependencyObject obj,
                        DependencyPropertyChangedEventArgs args)
        {
            (obj as PolylineArc).PropertyChangedCallback(args);
        }

        void PropertyChangedCallback(DependencyPropertyChangedEventArgs args)
        {
            PointCollection points = polylineSeg.Points;
            polylineSeg.Points = null;
            points.Clear();

            FlattenArc(points, StartPoint, Point, Size.Width, Size.Height,
                       RotationAngle, IsLargeArc,
                       SweepDirection == SweepDirection.Counterclockwise, 1);

            if (points.Count > 0)
                pathFig.StartPoint = points[0];

            polylineSeg.Points = points;
            InvalidateMeasure();
        }

        // CLR properties
        public Point StartPoint
        {
            set { SetValue(StartPointProperty, value); }
            get { return (Point)GetValue(StartPointProperty); }
        }

        public Point Point
        {
            set { SetValue(PointProperty, value); }
            get { return (Point)GetValue(PointProperty); }
        }

        public Size Size
        {
            set { SetValue(SizeProperty, value); }
            get { return (Size)GetValue(SizeProperty); }
        }

        public SweepDirection SweepDirection
        {
            set { SetValue(SweepDirectionProperty, value); }
            get { return (SweepDirection)GetValue(SweepDirectionProperty); }
        }

        public bool IsLargeArc
        {
            set { SetValue(IsLargeArcProperty, value); }
            get { return (bool)GetValue(IsLargeArcProperty); }
        }

        public double RotationAngle
        {
            set { SetValue(RotationAngleProperty, value); }
            get { return (double)GetValue(RotationAngleProperty); }
        }

        // Constructor
        public PolylineArc()
        {
            pathFig.Segments.Add(polylineSeg);
            pathGeo.Figures.Add(pathFig);
        }

        // Required override
        protected override Geometry DefiningGeometry
        {
            get { return pathGeo; }
        }

        void FlattenArc(IList<Point> points, Point pt1, Point pt2,
                double radiusX, double radiusY, double angleRotation,
                bool isLargeArc, bool isCounterclockwise, 
                double tolerance)
        {
            // Adjust for different radii and rotation angle
            Matrix matx = new Matrix();
            matx.Rotate(-angleRotation);
            matx.Scale(radiusY / radiusX, 1);
            pt1 = matx.Transform(pt1);
            pt2 = matx.Transform(pt2);

            // Get info about chord that connects both points
            Point midPoint = new Point((pt1.X + pt2.X) / 2, (pt1.Y + pt2.Y) / 2);
            Vector vect = pt2 - pt1;
            double halfChord = vect.Length / 2;

            // Get vector from chord to center
            Vector vectRotated;

            // (comparing two Booleans here!)
            if (isLargeArc == isCounterclockwise)
                vectRotated = new Vector(-vect.Y, vect.X);
            else
                vectRotated = new Vector(vect.Y, -vect.X);

            vectRotated.Normalize();

            // Distance from chord to center 
            double centerDistance = Math.Sqrt(radiusY * radiusY - halfChord * halfChord);

            // Calculate center point
            Point center = midPoint + centerDistance * vectRotated;

            // Get angles from center to the two points
            double angle1 = Math.Atan2(pt1.Y - center.Y, pt1.X - center.X);
            double angle2 = Math.Atan2(pt2.Y - center.Y, pt2.X - center.X);

            // (another comparison of two Booleans!)
            if (isLargeArc == (Math.Abs(angle2 - angle1) < Math.PI))
            {
                if (angle1 < angle2)
                    angle1 += 2 * Math.PI;
                else
                    angle2 += 2 * Math.PI;
            }

            // Invert matrix for final point calculation
            matx.Invert();

            // Calculate number of points for polyline approximation
            int max = (int)((4 * (radiusX + radiusY) * Math.Abs(angle2 - angle1) / (2 * Math.PI)) / tolerance);

            // Loop through the points
            for (int i = 0; i <= max; i++)
            {
                double angle = ((max - i) * angle1 + i * angle2) / max;
                double x = center.X + radiusY * Math.Cos(angle);
                double y = center.Y + radiusY * Math.Sin(angle);

                // Transform the point back
                Point pt = matx.Transform(new Point(x, y));
                points.Add(pt);
            }
        }
    }
}